#!/usr/bin/env python3
#===============================================================================
# pileups.py
#===============================================================================

# Imports ======================================================================

import argparse
import socket
import sys
import funcgenom




# Functions ====================================================================

def generate_counts(iterable, mode='cr', header=False):
    """Generate allele counts from a pileup file
    
    Parameters
    ----------
    iterable : iterable
        An iterable.
    mode : str
        Mode for producing allele counts. This should be a string of characters
        chosen from "a", "c", and/or "r". It defines which counts are included
        and in what order - e.g. "cr" means "coverage and ref count", while
        "cra" means "coverage, ref count, and alt count."
    header : bool
        If True, yield a header first.
    
    Yields
    ------
    tuple
        A row containing chromosome, position, and counts
    """

    if not set(mode) <= {'a', 'c', 'r'}:
        raise Exception('mode arg must include only "a", "c", and/or "r".')
    if header:
        header_dict = {'a': 'alt_count', 'c': 'coverage', 'r': 'ref_count'}
        yield ('chr', 'pos') + tuple(header_dict[m] for m in mode)
    for item in iterable:
        chromosome, position, _, coverage, read_bases, _ = (
            item.split() if isinstance(item, str) else item
        )
        ref_count = read_bases.replace(',', '.').count('.')
        count_dict = {
            'a': int(coverage) - ref_count,
            'c': int(coverage),
            'r': ref_count
        }
        yield (chromosome, int(position)) + tuple(count_dict[m] for m in mode)


def count_ref_alleles(variant, *traits):
    """Count reference allels for a variant
    
    Parameters
    ----------
    variant : a Variant as from funcgenom
        the variant for which alleles should be counted
    *traits : str
        the traits for which alleles should be counted
    

    Returns
    -------
    int
        the reference allele count
    """

    return (
        ''.join(variant.traits[trait]['alleles'] for trait in traits)
        .replace(',', '.')
        .count('.')
    )


def merge(*file_paths, alleles=True, count=False, reference=True, header=None):

    """Merge a group of pileup files on disk into a single pileup
    
    Parameters
    ----------
    *file_paths : str
        paths to pileup files
    alleles : bool
        include raw alleles in output if True
    count : bool
        include reference allele count in output if True
    reference : bool
        include reference allele in output if True
    header : iterable
        provide a header for the output
    
    Yields
    -------
    tuple
        a row of the merged pileup file
    """

    genome = funcgenom.Genome()
    for file_path in file_paths:
        genome.load_variants(
            file_path,
            add_header=('chr', 'pos', 'ref', 'coverage', 'alleles', 'qual'),
            traits={
                file_paths.index(file_path):
                    {'ref': 2, 'coverage': 3, 'alleles': 4, 'qual': 5}
            }
        )
        genome.sort_variants()
        genome.resolve_duplicate_variants()
    if header:
        yield header
    for variant in genome.variants():
        indices = tuple(variant.traits.keys())
        yield (
            ('chr{}'.format(variant.chromosome), str(variant.position))
            + reference * (variant.traits[indices[0]]['ref'].casefold(),)
            + (str(int(sum(variant.traits[i]['coverage'] for i in indices))),)
            + alleles * (
                ''.join(variant.traits[i]['alleles'] for i in indices),
                ''.join(str(variant.traits[i]['qual']) for i in indices)
            )
            + count * (str(count_ref_alleles(variant, *indices)),)
        )



#!/usr/bin/env python3
#===============================================================================
# pileups_count.py
#===============================================================================

"""Convert mpileup format to allele counts"""




# Imports ======================================================================

import argparse
import json
import os
import socket
import sys
import tempfile

from pileups.pileups import generate_counts
from pileups.allelecounts import join_as_dict





# Functions ====================================================================

def construct_counts_dict(
    pileup_file_path,
    mode='cr',
    header=True,
    temp_dir=None
):
    with tempfile.NamedTemporaryFile(dir=temp_dir) as temp_counts:
        temp_counts_name = temp_counts.name
    with open(pileup_file_path, 'r') as f, open(temp_counts_name, 'w') as g:
        g.write(
            '\n'.join(
                '\t'.join(str(item) for item in row)
                for row in generate_counts(
                    f, mode=mode, header=header
                )
            )
            + '\n'
        )
    counts = allelecounts.join_as_dict(temp_counts_name)
    os.remove(temp_counts_name)
    return counts


def print_tabular(pileup_file_path, mode='cr', header=True):
    with open(pileup_file_path, 'r') as f:
        for row in pileup.generate_counts(
            f,
            mode=mode,
            header=header
        ):
            print('\t'.join(str(item) for item in row))


def main(args):
    if args.json:
        print(
            json.dumps(
                construct_counts_dict(
                    args.pileup,
                    mode=args.mode,
                    header=not args.no_header,
                    temp_dir=args.tmp_dir
                )
            )
        )
    else:
        print_tabular(args.pileup, mode=args.mode, header=not args.no_header)


def parse_arguments():
    parser = argparse.ArgumentParser(
        description="convert a pileup to allele counts"
    )
    parser.add_argument(
        'pileup',
        metavar='<path/to/file.pileup>',
        help='path to pileup'
    )
    parser.add_argument(
        '--json',
        action='store_true',
        help='emit counts in JSON format'
    )
    parser.add_argument(
        '--no-header',
        action='store_true',
        help='do not include a header'
    )
    parser.add_argument(
        '--mode',
        metavar='<mode>',
        default='cr',
        help=(
            'string defining mode for counts, default is coverage and ref '
            'count [cr]'
        )
    )
    parser.add_argument(
        '--tmp-dir',
        metavar='<temp/file/dir/>',
        help='directory for temporary files'
    )
    return parser.parse_args()




# Execute ======================================================================

args = parse_arguments()
main(args)



#!/usr/bin/env python3
#===============================================================================
# pileups_merge.py
#===============================================================================

"""Merge pileup files"""




# Imports ======================================================================

import argparse
import socket
import sys

from pileups.pileups import merge




# Functions ====================================================================

def parse_arguments():
    parser = argparse.ArgumentParser(description='Merge pileup files')
    parser.add_argument('pileup', nargs='+', help='Paths to pileup files')
    parser.add_argument(
        '--alleles',
        action='store_true',
        help='Print pileup alleles and quality scores'
    )
    parser.add_argument(
        '--count',
        action='store_true',
        help='Print reference allele counts'
    )
    parser.add_argument(
        '--reference',
        action='store_true',
        help='Print reference alleles'
    )
    parser.add_argument('--header', help='Add a header to the output')
    args = parser.parse_args()
    if not args.count:
        args.alleles = True
    if args.alleles:
        args.reference = True
    return args


def main():
    args = parse_arguments()
    for row in merge(
        *args.pileup,
        alleles=args.alleles,
        count=args.count,
        reference=args.reference,
        header=(args.header.split() if args.header else None)
    ):
        print('\t'.join(row), end='\n')



#!/usr/bin/env python3
#===============================================================================
# allelecounts.py
#===============================================================================

"""Functions for handling allele count data"""




# Imports ======================================================================

import os.path
import socket
import sys
import funcgenom




# Functions ====================================================================

def construct_dict(genome, *dataset_names):
    """
    Construct an allele counts dict from a genome object

    Parameters
    ----------
    genome : funcgenom.Genome()
        a genome as from funcgenom
    *dataset_names : str
        names for the datasets represented by the genome
    
    Returns
    -------
    dict
        a dictionary whose values are tables containing the coordinates and
        allele counts
    """

    return {
        'coordinates': dict(
            zip(
                ('chr', 'pos'),
                (
                    list(z) for z in zip(
                        *((v.chromosome, v.position) for v in genome.variants())
                    )
                )
            )
        ),
        **{
            dataset_name: dict(
                zip(
                    ('coverage', 'ref_count'),
                    (
                        list(z) for z in zip(
                            *(
                                (
                                    v.traits.get(dataset_name, {}).get(
                                        'coverage'
                                    ),
                                    v.traits.get(dataset_name, {}).get(
                                        'ref_count'
                                    )
                                )
                                for v in genome.variants()
                            )
                        )
                    )
                )
            )
            for dataset_name in dataset_names
        }
    }


def join_as_dict(
    *counts_file_paths,
    names=None,
    coverage_index=2,
    ref_count_index=3,
    **kwargs
):
    """
    Read allele counts files and join them as a dictionary

    Parameters
    ----------
    counts_file_paths : str
        paths to allele counts files
    dataset_names : iterable
        names corresponding to counts file paths
    coverage_index : int
        column index for coverage in counts files
    ref_count_index : int
        column index for ref count in counts files
    **kwargs
        other parameters for `Genome.load_variants`
    
    Returns
    -------
    dict
        a dictionary whose values are tables containing the coordinates and
        allele counts
    """
    names = names if names else tuple(
        os.path.basename(path).replace('.pileup', '')
        for path in counts_file_paths
    )
    genome = funcgenom.Genome()
    for counts_file_path, name in zip(counts_file_paths, names):
        genome.load_variants(
            counts_file_path,
            traits={
                name: {
                    'coverage': coverage_index,
                    'ref_count': ref_count_index
                }
            },
            **kwargs
        )
        genome.sort_variants()
        genome.resolve_duplicate_variants()
    return construct_dict(genome, *names)
